/*
 * aligner.h
 *
 *  Created on: 2010/09/28
 *      Author: shu
 */

#ifndef ALIGNER_H_
#define ALIGNER_H_

#include "encoder.h"
#include "edit_blocks.h"
#include "sequence_type.h"
#include "score_matrix.h"
#include "score_matrix_reader.h"
#include "aligner_parameters.h"
#include "query.h"
#include <stdint.h>
#include <vector>
#include <string>

using namespace std;

class Queries;
class Db;

typedef struct {
  uint32_t queries_offset;
  uint32_t db_offset;
} Coordinate;

typedef struct {
  int score;
  Coordinate start;
  Coordinate end;
  Coordinate hit;
} Hit;

typedef struct {
  uint32_t db_id;
  uint32_t subject_id;
  string subject_name;
  int score;
  uint32_t alignment_length;
  uint32_t mismatches;
  uint32_t gap_openings;
  uint32_t gap_extendings;
  Coordinate start;
  Coordinate end;
  Coordinate hit;
} Alignment;

class Aligner {
public:
  Aligner();
  virtual ~Aligner();

  virtual void Run(AlignerParameters &parameters);

protected:
  uint32_t kNextQueryIdIncrement_;
  uint32_t kNextAlignmentIdIncrement_;
  
  typedef struct {
    uint32_t position;
    uint32_t index;
  }PositionEntry;

  typedef struct {
    int thread_id;
    Aligner *aligner;
    AlignerParameters *parameters;
    Queries *queries;
    Db *db;
    vector< vector<Alignment> > *best_alignments;
  }PresearchThreadArguments;

  typedef struct {
    int thread_id;
    Aligner *aligner;
    AlignerParameters *parameters;
    Queries *queries;
    Db *db;
    vector< vector<Alignment> > *best_alignments;
    vector< vector< pair<uint32_t, uint32_t> > > *alignments_id_list;
  }TraceBackThreadArguments;

  class QueryPositionEntryComp {
  public:
    bool operator()(const PositionEntry &a1, const PositionEntry &a2) const {
      return a1.position <  a2.position;
    }
  };

  class HitComp {
  public:
    bool operator()(const Hit &a1, const Hit &a2) const {
      return (Aligner::AlignmentComp(a1.score, a1.start, a1.end, a2.score, a2.start, a2.end) > 0);
    }
  };

  class FinalAlignmentComp {
  public:
    bool operator()(const Alignment &a1, const Alignment &a2) const {
      return (Aligner::AlignmentComp(a1.score, a1.start, a1.end, a2.score, a2.start, a2.end) > 0);
    }
  };
  
  class HitStartComp {
  public:
    bool operator()(const Hit &a1, const Hit &a2) const {
      return a1.start.queries_offset <  a2.start.queries_offset;
    }
  };



  static int AlignmentComp(int a_score, Coordinate a_start, Coordinate a_end, int b_score, Coordinate b_start, Coordinate b_end);
  static bool Overlapped(Coordinate a_start, Coordinate a_end, Coordinate b_start, Coordinate b_end);
  

  static void *RunPresearchThread(void *arg);
  static void *RunTraceBackThread(void *arg);

  uint32_t GetNextQueryId(Queries &queries);
  void ResetNextQueryId();
  uint32_t GetNextAlignmentId(vector< pair<uint32_t, uint32_t> > &alignments);
  void ResetNextAlignmentId();
  void Finished(const AlignerParameters &parameters);
  int GetFinished();
  void ResetFinished();

  virtual void Presearch(AlignerParameters &parameters, Queries &queries, Db &db, vector< vector<Alignment> > &best_alignments);
  virtual void PresearchThread(int thread_id, AlignerParameters &parameters, Queries &queries, Db &db, vector< vector<Alignment> > &best_alignments);
  virtual void FillAlignmentData(Query &query, Db &db, vector<EditBlocks::EditOpType> &edits, Alignment *alignment);
  virtual void Merge(const AlignerParameters &parameters, vector<Hit> &hits, Db &db, vector< vector<PositionEntry> > &start_lists, vector< vector<PositionEntry> > &end_lists, vector<Alignment> &best_alignments);
  virtual void TraceBack(AlignerParameters &parameters, vector< vector<Alignment> > &best_alignments, Queries &queries, Db &db);
  virtual void TraceBackThread(int thread_id, AlignerParameters &parameters, vector< vector<Alignment> > &best_alignments, vector< vector< pair<uint32_t, uint32_t> > > &alignments_id_list, Queries &queries, Db &db);
  virtual void WriteOutput(const AlignerParameters &parameters, ostream &out, vector< vector<Alignment> > &best_alignments, Queries &queries, Db &db);
  

  volatile int finished_; 
  volatile uint32_t next_query_id_;
  volatile uint32_t next_alignment_id_;
  
};

#endif /* ALIGNER_H_ */
