/*
 * bit_vector.h
 *
 *  Created on: 2009/11/13
 *      Author: shu
 */

#ifndef BIT_VECTOR_H_
#define BIT_VECTOR_H_

#include <stdint.h>

class BitVector {
public:
  static const uint32_t kLogBlockSize = 5;
  static const uint32_t kBlockSize = (1 << kLogBlockSize);

  BitVector
  (
      uint32_t size // the number of bits
  );

  BitVector
  (
      uint32_t *blocks,
      const uint32_t size // the number of bits
  )
  {
    blocks_ = blocks;
    size_ = size;
    blocks_size_ = (size_ + kBlockSize - 1)/kBlockSize;
  }

  virtual ~BitVector
  (
      // no parameter
  );

  void SetBit
  (
      const uint32_t position,
      const uint32_t bit
  )
  {
    const uint32_t block_position = position/kBlockSize;
    const uint32_t offset = position % kBlockSize;
    if (bit == 1) {
      blocks_[block_position] |= (1U << offset);
    } else {
      blocks_[block_position] &= (~(1U << offset));
    }
  }

  // set 1
  void SetBit
  (
      const uint32_t position
  )
  {
    SetBit(position, 1);
  }

  void SetBits
  (
      const uint32_t position, // start set position
      const uint32_t length, // bits length. max 32
      const uint32_t bits
  )
  {
    for (uint32_t i = 0; i < length; ++i) {
      SetBit(position + i, (bits >> (length - 1 -i))&1);
    }
  }

  uint32_t GetBit
  (
      const uint32_t position // bit position
  ) const
  {
    const uint32_t block_position = position >> kLogBlockSize;
    const uint32_t offset = position - (block_position << kLogBlockSize);
    return (blocks_[block_position]>> offset) & 1;
  }

  uint32_t GetBits
  (
      const uint32_t position, // start position
      const uint32_t length // bits length
  ) const
  {
    uint32_t bits = 0;
    for (uint32_t i = 0; i < length; ++i) {
      bits <<= 1;
      bits += GetBit(position + i);
    }
    return bits;
  }

  uint32_t GetBlock
  (
      const uint32_t position // block position
  )
  {
    return blocks_[position];
  }

  uint32_t *GetBlocks
  (
      // no parameter
  )
  {
    return blocks_;
  }

  uint32_t GetSize
  (
      // no parameter
  ) const
  {
    return size_;
  }

  uint32_t GetBlocksSize
  (
      // no parameter
  )
  {
    return blocks_size_;
  }

  uint32_t operator []
  (
      const uint32_t position
  ) const
  {
    return GetBit(position);
  }

private:
  uint32_t size_;
  uint32_t blocks_size_;
  uint32_t *blocks_;
};

#endif /* BIT_VECTOR_H_ */
