/*
 * succinct_bit_vector.h
 *
 *  Created on: 2009/09/07
 *      Author: shu
 */

#ifndef SUCCINCT_BIT_VECTOR_H_
#define SUCCINCT_BIT_VECTOR_H_

#include <stdint.h>
#include "bit_vector.h"

class SuccinctBitVector {
public:
  static const uint32_t kLogLevelA = 8;
  static const uint32_t kLogLevelB = 5;
  static const uint32_t kLevelA = (1 << kLogLevelA);
  static const uint32_t kLevelB = (1 << kLogLevelB);

  SuccinctBitVector
  (
      BitVector &bit_vector
  );

  SuccinctBitVector
  (
      BitVector &bit_vector,
      uint32_t level_a_size,
      uint32_t level_b_size,
      uint32_t level_a[],
      uint8_t level_b[]
  );


  virtual ~SuccinctBitVector
  (
      // no parameter
  );

  uint32_t Rank1
  (
      uint32_t position // position start is 1. not 0
  )
  {
    uint32_t block_position = position >> BitVector::kLogBlockSize;
    return level_a_[position >> kLogLevelA] +
    level_b_[position >> kLogLevelB] +
      PopCount(blocks_[block_position] & ((1 << (position - (block_position << BitVector::kLogBlockSize))) - 1));
  }

  uint32_t Rank
  (
      uint32_t position,
      uint32_t bit
  )
  {
    ++position;
    if (bit == 1) {
      return Rank1(position);
    } else {
      return position - Rank1(position);
    }
  }

  uint32_t GetLevelASize
  (
      // no parameter
  )
  {
    return level_a_size_;
  }

  uint32_t GetLevelBSize
  (
      // no parameter
  )
  {
    return level_b_size_;
  }

  uint32_t* GetLevelA
  (
      // no parameter
  )
  {
    return level_a_;
  }

  uint8_t* GetLevelB
  (
      // no parameter
  )
  {
    return level_b_;
  }

private:
  static const unsigned char kPopCountArray[];

  const uint32_t *blocks_;

  uint32_t level_a_size_;
  uint32_t level_b_size_;

  uint32_t *level_a_;
  uint8_t *level_b_;

  uint32_t PopCount(uint32_t bits) {
    return kPopCountArray[(bits >> 24) & 0xFF] +
    kPopCountArray[(bits >> 16) & 0xFF] +
    kPopCountArray[(bits >> 8 ) & 0xFF] +
    kPopCountArray[(bits      ) & 0xFF];
  }

};

#endif /* SUCCINCT_BIT_VECTOR_H_ */
